"""Utility Modules."""


from typing import Any, Callable, Dict, List, Optional, Sequence, Type, TypeVar, Union

import jax
import jax.numpy as jnp

from .base import PaxKind, allow_mutation
from .module import Module

T = TypeVar("T", bound=Module)
O = TypeVar("O")


class ParameterModule(Module):
    """A PAX module that registers attributes as parameters by default."""

    def __new__(cls: Type[T], *args, **kwargs) -> T:
        """Set the default kind to ``PaxKind.PARAMETER``."""
        module = super().__new__(cls, *args, **kwargs)
        with allow_mutation(module):
            module._update_default_kind(PaxKind.PARAMETER)
        return module


class StateModule(Module):
    """A PAX module that registers attributes as states by default."""

    def __new__(cls: Type[T], *args, **kwargs) -> T:
        """Set the default kind to ``PaxKind.STATE``."""
        module = super().__new__(cls, *args, **kwargs)
        with allow_mutation(module):
            module._update_default_kind(PaxKind.STATE)
        return module


class LazyModule(Module):
    """A lazy module is a module that only creates submodules when needed.


    Example:

    >>> from dataclasses import dataclass
    >>> @dataclass
    ... class MLP(pax.experimental.LazyModule):
    ...     features: list
    ...
    ...     def __call__(self, x):
    ...         sizes = zip(self.features[:-1], self.features[1:])
    ...         for i, (in_dim, out_dim) in enumerate(sizes):
    ...             fc = self.get_or_create(f"fc_{i}", lambda: pax.nn.Linear(in_dim, out_dim))
    ...             x = jax.nn.relu(fc(x))
    ...         return x
    ...
    ...
    >>> mlp, _ = MLP([1, 2, 3]) % jnp.ones((1, 1))
    >>> print(mlp.summary())
    MLP(features=[1, 2, 3])
    ├── Linear(in_dim=1, out_dim=2, with_bias=True)
    └── Linear(in_dim=2, out_dim=3, with_bias=True)
    """

    def get_or_create(self, name, create_fn: Callable[[], T], kind=PaxKind.MODULE) -> T:
        """Create and register a new attribute when it is not exist.

        Return the attribute.
        """
        if hasattr(self, name):
            value = getattr(self, name)
        else:
            assert callable(create_fn), "Expect a callable function"
            value = create_fn()
            self.register_subtree(name, value, kind)
        return value

    def get_or_create_parameter(self, name, create_fn: Callable[[], O]) -> O:
        """Get or create a trainable parameter."""
        return self.get_or_create(name, create_fn=create_fn, kind=PaxKind.PARAMETER)

    def get_or_create_state(self, name, create_fn: Callable[[], O]) -> O:
        """Get or create a non-trainable state."""
        return self.get_or_create(name, create_fn=create_fn, kind=PaxKind.STATE)


class Lambda(Module):
    """Convert a function to a module.

    Example:

    >>> net = pax.nn.Lambda(jax.nn.relu)
    >>> print(net.summary())
    x => relu(x)
    >>> y = net(jnp.array(-1))
    >>> y
    DeviceArray(0, dtype=int32, weak_type=True)
    """

    func: Callable

    def __init__(self, func: Callable, name: Optional[str] = None):
        super().__init__(name=name)
        self.func = func

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)

    def __repr__(self) -> str:
        if self.name is not None:
            return super().__repr__()
        else:
            return f"{self.__class__.__qualname__}({self.func.__name__})"

    def summary(self, return_list: bool = False) -> Union[str, List[str]]:
        if self.name is not None:
            name = self.name
        elif isinstance(self.func, jax.custom_jvp) and hasattr(self.func, "fun"):
            if hasattr(self.func.fun, "__name__"):
                name = self.func.fun.__name__
            else:
                name = f"{self.func.fun}"
        elif hasattr(self.func, "__name__"):
            name = self.func.__name__
        else:
            name = f"{self.func}"
        output = f"x => {name}(x)"
        return [output] if return_list else output


class Flattener(Module):
    """Flatten PAX modules for better performance.

    Example:

    >>> net = pax.nn.Linear(3, 3)
    >>> opt = opax.adam(1e-3)(net.parameters())
    >>> flat_mods = pax.experimental.Flattener(model=net, optimizer=opt)
    >>> net, opt = flat_mods.model, flat_mods.optimizer
    >>> print(net.summary())
    Linear(in_dim=3, out_dim=3, with_bias=True)
    >>> print(opt.summary())
    chain.<locals>.Chain
    ├── scale_by_adam.<locals>.ScaleByAdam
    └── scale.<locals>.Scale
    """

    treedef_dict: Dict[str, Any]
    leaves_dict: Dict[str, Sequence[jnp.ndarray]]

    def __init__(self, **kwargs):
        """Create a new flattener."""
        super().__init__()
        self.treedef_dict = {}
        self.leaves_dict = {}
        self.set_attribute_kind(leaves_dict=PaxKind.STATE)

        for name, value in kwargs.items():
            leaves, treedef = jax.tree_flatten(value)
            self.treedef_dict[name] = treedef
            self.leaves_dict[name] = leaves

    def __getattr__(self, name: str) -> Any:
        if name in self.treedef_dict:
            treedef = self.treedef_dict[name]
            leaves = self.leaves_dict[name]
            value = jax.tree_unflatten(treedef, leaves)
            return value
        else:
            raise AttributeError()

    def update(self: T, **kwargs) -> T:
        """Update the flattener.

        Example:

        >>> net = pax.nn.Linear(3, 3)
        >>> flats = pax.experimental.Flattener(net=net)
        >>> flats = flats.update(net=pax.nn.Linear(4, 4))
        >>> print(flats.net.summary())
        Linear(in_dim=4, out_dim=4, with_bias=True)
        """
        new_self = self.copy()
        for name, value in kwargs.items():
            leaves, treedef = jax.tree_flatten(value)
            new_self.treedef_dict[name] = treedef
            new_self.leaves_dict[name] = leaves
        return new_self

    def parameters(self: T) -> T:
        """Raise an error.

        Need to reconstruct the original module before getting parameters.
        """

        raise ValueError(
            "A flattener only stores ndarray leaves as non-trainable states.\n"
            "Reconstruct the original module before getting parameters."
        )



"""Dropout module."""

from typing import Optional

import jax
import jax.numpy as jnp

from ..core import StateModule
from ..core.rng import KeyArray, next_rng_key


def dropout(rng_key: KeyArray, dropout_rate: float, x: jnp.ndarray) -> jnp.ndarray:
    """Dropout input `x` randomly.

    Scaling the input by ``1 / (1-dropout_rate)`` makes ``E[output] = input``.
    """
    assert 0 <= dropout_rate < 1.0

    if dropout_rate == 0.0:
        return x
    else:
        mask = jax.random.bernoulli(rng_key, dropout_rate, shape=x.shape)
        x = jnp.where(mask, 0.0, x / (1.0 - dropout_rate))
        return x


class Dropout(StateModule):
    """A Dropout Module.

    Dropout module stores an internal state ``rng_key``.
    It refreshes ``rng_key`` whenever a forward pass is executed.
    """

    rng_key: KeyArray
    dropout_rate: float

    def __init__(self, dropout_rate: float, *, name: Optional[str] = None):
        """Create a dropout module.

        Arguments:
            dropout_rate: the probability of dropping an element.
            name: the module name.
        """
        super().__init__(name=name)
        assert 0 <= dropout_rate < 1.0

        self.dropout_rate = dropout_rate
        self.rng_key = next_rng_key()

    def __call__(self, x):
        """Dropout `x` randomly.

        Return the input `x` if in `eval` mode or `dropout_rate=0`.
        """

        if self.training and self.dropout_rate > 0:
            self.rng_key, rng_key = jax.random.split(self.rng_key)
            return dropout(rng_key, self.dropout_rate, x)
        else:
            return x

    def __repr__(self):
        return super()._repr({"dropout_rate": self.dropout_rate})



"""EMA module."""

from typing import Any, Optional

import jax
import jax.numpy as jnp

from ..core import StateModule


class EMA(StateModule):
    """Exponential Moving Average (EMA) Module"""

    averages: Any
    decay_rate: float
    debias: Optional[jnp.ndarray]

    def __init__(self, initial_value, decay_rate: float, debias: bool = False):
        """Create a new EMA module.

        Arguments:
            initial_value: the initial value.
            decay_rate: the decay rate.
            debias: ignore the initial value to avoid biased estimates.
        """

        super().__init__()
        self.averages = initial_value
        self.decay_rate = decay_rate
        if debias:
            # avoid integer ndarray for `jax.grad` convenience,
            # e.g., no need to pass `allow_int=True` to `jax.grad`.
            self.debias = jnp.array(0.0)
        else:
            self.debias = None

    def __call__(self, xs):
        """Return the ema of `xs`. Also, update internal states."""

        if self.training:
            if self.debias is not None:
                cond = self.debias > 0
                self.averages = jax.tree_map(
                    lambda a, x: jnp.where(cond, a, x), self.averages, xs
                )

                self.debias = jnp.array(1.0)

            self.averages = jax.tree_map(
                lambda a, x: a * self.decay_rate + x * (1 - self.decay_rate),
                self.averages,
                xs,
            )

        return self.averages



"""Useful functions."""

import functools
from typing import Any, Callable, Dict, Tuple, TypeVar, Union

import jax
import jax.numpy as jnp
import jmp

from .core import (
    Module,
    apply_mp_policy,
    module_and_value,
    pure,
    select_parameters,
    update_parameters,
)
from .nn import (
    BatchNorm1D,
    BatchNorm2D,
    Conv1D,
    Conv1DTranspose,
    Conv2D,
    Conv2DTranspose,
    GroupNorm,
    LayerNorm,
    Linear,
)

GradientTransformation = Module
T = TypeVar("T", bound=Module)
O = TypeVar("O", bound=GradientTransformation)
C = TypeVar("C")
K = TypeVar("K")


def grad(
    fun: Union[
        Callable[[T, Any], Tuple[jnp.ndarray, C]],
        Callable[[T, Any, Any], Tuple[jnp.ndarray, C]],
        Callable[[T, Any, Any, Any], Tuple[jnp.ndarray, C]],
        Callable[..., Tuple[jnp.ndarray, C]],
    ],
    *args,
    **kwargs
) -> Callable[..., Tuple[T, C]]:
    """Compute gradient with respect to trainable parameters of the first argument.

    Example:

    >>> @pax.pure
    ... def loss_fn(model: pax.nn.Linear, x, y):
    ...     y_hat = model(x)
    ...     loss = jnp.mean(jnp.square(y - y_hat))
    ...     return loss, (loss, model)
    ...
    >>> grad_fn = pax.grad(loss_fn, has_aux=True)
    >>> net = pax.nn.Linear(1, 1)
    >>> x = jnp.ones((3, 1))
    >>> grads, (loss, net) = grad_fn(net, x, x)
    >>> print(jax.tree_leaves(grads))
    [DeviceArray([[-1.4543512]], dtype=float32), DeviceArray([-1.4543512], dtype=float32)]
    """

    def _fun(params: T, mod: T, *args, **kwargs):
        mod = mod.update_parameters(params.parameters())
        out = fun(mod, *args, **kwargs)
        return out

    grad_fn_ = jax.grad(_fun, *args, **kwargs)

    def grad_fn(mod: T, *args, **kwargs) -> Tuple[T, C]:
        if not isinstance(mod, Module):
            raise ValueError("Expecting a PAX's Module at the first argument.")

        out = grad_fn_(mod.parameters(), mod, *args, **kwargs)
        return out

    return grad_fn


def value_and_grad(func: Callable[..., K], has_aux=False):
    """A PAX-compatible version of jax.value_and_grad.

    This version computes gradients w.r.t. trainable parameters of a PAX module.
    """

    def func_with_params(params: T, module: T, *args, **kwargs):
        return func(module | params, *args, **kwargs)

    vag_fn = jax.value_and_grad(func_with_params, has_aux=has_aux)

    @functools.wraps(vag_fn)
    def new_vag_fn(module: T, *args, **kwargs) -> Tuple[K, T]:
        return vag_fn(module.parameters(), module, *args, **kwargs)

    return new_vag_fn


def build_update_fn(loss_fn, *, scan_mode: bool = False):
    """Build a simple update function.

    *Note*: The output of ``loss_fn`` must be ``(loss, (aux, model))``.

    Arguments:
        loss_fn: The loss function.
        scan_mode: If true, use `(model, optimizer)` as a single argument.

    Example:

    >>> def mse_loss(model, x, y):
    ...     y_hat = model(x)
    ...     loss = jnp.mean(jnp.square(y - y_hat))
    ...     return loss, (loss, model)
    ...
    >>> update_fn = pax.utils.build_update_fn(mse_loss)
    >>> net = pax.nn.Linear(2, 2)
    >>> optimizer = opax.adam(1e-4)(net.parameters())
    >>> x = jnp.ones((32, 2))
    >>> y = jnp.zeros((32, 2))
    >>> net, optimizer, loss = update_fn(net, optimizer, x, y)
    """

    # pylint: disable=import-outside-toplevel
    from opax import apply_updates, transform_gradients

    def _update_fn(model: T, optimizer: O, *inputs, **kwinputs) -> Tuple[T, O, Any]:
        """An update function.

        Note that: ``model`` and ``optimizer`` have internal states.
        We have to return them in the output as jax transformations
        (e.g., ``jax.grad`` and ``jax.jit``) requires pure functions.


        Arguments:
            model_and_optimizer: (a callable pax.Module, an optimizer),
            inputs: input batch.

        Returns:
            model_and_optimizer: updated (model, optimizer),
            aux: the aux info.
        """

        assert isinstance(model, Module)
        assert isinstance(optimizer, Module)

        model_treedef = jax.tree_structure(model)
        grads, (aux, model) = grad(loss_fn, has_aux=True)(model, *inputs, **kwinputs)
        if jax.tree_structure(model) != model_treedef:
            raise ValueError("Expecting an updated model in the auxiliary output.")

        params = select_parameters(model)
        updates, optimizer = transform_gradients(grads, optimizer, params=params)
        params = apply_updates(params, updates=updates)
        model = update_parameters(model, params=params)
        return model, optimizer, aux

    def _update_fn_scan(
        model_and_optimizer: Union[C, Tuple[T, O]], *inputs, **kwinputs
    ) -> Tuple[C, Any]:
        model, optimizer = model_and_optimizer
        model, optimizer, aux = _update_fn(model, optimizer, *inputs, **kwinputs)
        return (model, optimizer), aux

    return _update_fn_scan if scan_mode else _update_fn


def scan(func, init, xs, length=None, unroll: int = 1, time_major=True):
    """``jax.lax.scan`` with an additional ``time_major=False`` mode.


    The semantics of ``scan`` are given roughly by this Python implementation:

    >>> def scan(f, init, xs, length=None):
    ...     if xs is None:
    ...         xs = [None] * length
    ...     carry = init
    ...     ys = []
    ...     for x in xs:
    ...         carry, y = f(carry, x)
    ...         ys.append(y)
    ...     return carry, np.stack(ys)
    """
    if time_major:
        # data format: TN...
        return jax.lax.scan(func, init, xs, length=length, unroll=unroll)
    else:
        # data format: NT...
        if xs is not None:
            xs = jnp.swapaxes(xs, 0, 1)  # swap batch and time axes
        state, output = jax.lax.scan(func, init, xs, length=length, unroll=unroll)
        output = jnp.swapaxes(output, 0, 1)  # restore to NT...
        return state, output


def apply_scaled_gradients(model: T, optimizer: Callable, loss_scale, grads: T):
    """Update model, optimizer and loss scale.

    Example:

    >>> import jmp
    >>> from pax.experimental import apply_scaled_gradients
    >>> net = pax.nn.Linear(2, 2)
    >>> opt = opax.adam(1e-4)(net.parameters())
    >>> loss_scale = jmp.DynamicLossScale(jmp.half_dtype()(2**15))
    >>> grads = net.parameters()
    >>> net, opt, loss_scale = apply_scaled_gradients(net, opt, loss_scale, grads)
    >>> print(loss_scale.loss_scale)
    32770.0
    """
    params = model.parameters()
    grads = loss_scale.unscale(grads)
    skip_nonfinite_updates = isinstance(loss_scale, jmp.DynamicLossScale)
    if skip_nonfinite_updates:
        grads_finite = jmp.all_finite(grads)
        loss_scale = loss_scale.adjust(grads_finite)
        new_optimizer, updates = module_and_value(optimizer)(grads, params)
        new_params = params.map(jax.lax.sub, updates)
        new_model = model.update_parameters(new_params)
        model, optimizer = jmp.select_tree(
            grads_finite,
            (new_model, new_optimizer),
            (model, optimizer),
        )
    else:
        optimizer, updates = module_and_value(optimizer)(grads, params)
        params = params.map(jax.lax.sub, updates)
        model = model.update_parameters(params)
    return model, optimizer, loss_scale


def default_mp_policy(module: T) -> T:
    """A default mixed precision policy.

    - Linear layers are in half precision.
    - Normalization layers are in full precision.

    Example:

    >>> net = pax.nn.Sequential(pax.nn.Linear(3, 3), pax.nn.BatchNorm1D(3))
    >>> net = net.apply(pax.experimental.default_mp_policy)
    >>> print(net.summary())
    Sequential
    ├── Linear(in_dim=3, out_dim=3, with_bias=True, mp_policy=FHF)
    └── BatchNorm1D(num_channels=3, ..., mp_policy=FFF)
    """
    half = jmp.half_dtype()
    full = jnp.float32
    linear_policy = jmp.Policy(compute_dtype=half, param_dtype=full, output_dtype=full)
    norm_policy = jmp.Policy(compute_dtype=full, param_dtype=full, output_dtype=full)
    linear_classes = (Linear, Conv1D, Conv2D, Conv1DTranspose, Conv2DTranspose)
    norm_classes = (BatchNorm1D, BatchNorm2D, GroupNorm, LayerNorm)

    if isinstance(module, linear_classes):
        return apply_mp_policy(module, mp_policy=linear_policy)
    elif isinstance(module, norm_classes):
        return apply_mp_policy(module, mp_policy=norm_policy)
    else:
        return module  # unchanged


def save_weights_to_dict(module: Module) -> Dict[str, Any]:
    """Save module weights to a dictionary.

    >>> net = pax.nn.Sequential(pax.nn.Linear(1, 2), jax.nn.relu, pax.nn.Linear(2, 3))
    >>> weights = pax.experimental.save_weights_to_dict(net)
    >>> weights
    {'modules': ({'weight': ..., 'bias': ...}, {}, {'weight':..., 'bias': ...})}
    """

    def _dict_or_array(v):
        return save_weights_to_dict(v) if isinstance(v, Module) else v

    out = {}
    for name in module.pax.name_to_kind:
        value = getattr(module, name)
        out[name] = jax.tree_map(
            _dict_or_array,
            value,
            is_leaf=lambda x: isinstance(x, Module),
        )

    return out


@pure
def load_weights_from_dict(module: T, state_dict: Dict[str, Any]) -> T:
    """Load module weights from a dictionary.

    >>> a = pax.nn.Sequential(pax.nn.Linear(1, 2), jax.nn.relu, pax.nn.Linear(2, 3))
    >>> weights = pax.experimental.save_weights_to_dict(a)
    >>> b = pax.nn.Sequential(pax.nn.Linear(1, 2), jax.nn.relu, pax.nn.Linear(2, 3))
    >>> b = pax.experimental.load_weights_from_dict(b, weights)
    >>> assert a == b
    """

    def _module_or_array(m, s):
        return load_weights_from_dict(m, s) if isinstance(m, Module) else s

    out = module
    for name in module.pax.name_to_kind:
        value = getattr(module, name)
        value = jax.tree_map(
            _module_or_array,
            value,
            state_dict[name],
            is_leaf=lambda x: isinstance(x, Module),
        )
        setattr(out, name, value)

    return out



"""Modules."""

from .attention import MultiHeadAttention
from .batch_norm import BatchNorm1D, BatchNorm2D
from .conv import Conv1D, Conv1DTranspose, Conv2D, Conv2DTranspose
from .dropout import Dropout
from .ema import EMA
from .embed import Embed
from .group_norm import GroupNorm
from .identity import Identity
from .lambda_module import Lambda
from .layer_norm import LayerNorm
from .linear import Linear
from .pool import avg_pool, max_pool
from .recurrent import GRU, LSTM, GRUState, LSTMState
from .rng_seq import RngSeq
from .sequential import Sequential



"""LayerNorm Module."""

# The implementation is almost identical to dm-haiku LayerNorm at:
# https://github.com/deepmind/dm-haiku/blob/main/haiku/_src/layer_norm.py
# deepmind/dm-haiku is licensed under the Apache License 2.0
#
# Differences:
#     1. We need to input ``num_channels``, the size of the last dimension,
#        to initialize scale/offset parameters.
#     2. We can input `rng_key` to seed the value of scale/offset parameters.

import collections
from typing import Callable, Optional, Sequence, Union

import jax
import jax.numpy as jnp
import numpy as np

from ..core import ParameterModule
from ..core.rng import KeyArray, next_rng_key


class LayerNorm(ParameterModule):
    """LayerNorm module.
    See: https://arxiv.org/abs/1607.06450.
    """

    scale: Optional[jnp.ndarray]
    offset: Optional[jnp.ndarray]

    def __init__(
        self,
        num_channels: int,
        axis: Union[int, Sequence[int], slice],
        create_scale: bool,
        create_offset: bool,
        eps: float = 1e-5,
        scale_init: Optional[Callable] = None,
        offset_init: Optional[Callable] = None,
        *,
        rng_key: Optional[KeyArray] = None,
        name: Optional[str] = None,
    ):
        jax.nn.initializers
        """Constructs a LayerNorm module.

        Arguments:
            num_channels: Integer, size of the last dimension. The data format is ``[N, ..., C]``.
            axis: Integer, list of integers, or slice indicating which axes to normalize over.
            create_scale: Bool, defines whether to create a trainable scale
              per channel applied after the normalization.
            create_offset: Bool, defines whether to create a trainable offset
              per channel applied after normalization and scaling.
            eps: Small epsilon to avoid division by zero variance.
              Defaults ``1e-5``, as in the paper and Sonnet.
            scale_init: Optional initializer for gain (aka scale). By default, one.
            offset_init: Optional initializer for bias (aka offset). By default, zero.
            rng_key: RNG key.
            name: module name.
        """
        super().__init__(name=name)
        if not create_scale and scale_init is not None:
            raise ValueError("Cannot set `scale_init` if `create_scale=False`.")
        if not create_offset and offset_init is not None:
            raise ValueError("Cannot set `offset_init` if `create_offset=False`.")

        if isinstance(axis, slice):
            self.axis = axis
        elif isinstance(axis, int):
            self.axis = (axis,)
        elif isinstance(axis, collections.abc.Iterable) and all(
            isinstance(ax, int) for ax in axis
        ):
            self.axis = tuple(axis)
        else:
            raise ValueError("`axis` should be an int, slice or iterable of ints.")

        self.eps = eps
        self.create_scale = create_scale
        self.create_offset = create_offset
        self.scale_init = scale_init or jax.nn.initializers.ones
        self.offset_init = offset_init or jax.nn.initializers.zeros
        self.num_channels = num_channels

        param_shape = [num_channels]
        rng_key = next_rng_key() if rng_key is None else rng_key
        rng1, rng2 = jax.random.split(rng_key)
        if create_scale:
            self.scale = self.scale_init(rng1, param_shape)
        else:
            self.scale = None
        if create_offset:
            self.offset = self.offset_init(rng2, param_shape)
        else:
            self.offset = None

    def __call__(
        self,
        inputs: jnp.ndarray,
        scale: Optional[jnp.ndarray] = None,
        offset: Optional[jnp.ndarray] = None,
    ) -> jnp.ndarray:
        """Returns normalized inputs.

        Arguments:
            inputs: An array, where the data format is ``[N, ..., C]``.
            scale: An array up to n-D. The shape of this tensor must be broadcastable
                to the shape of ``inputs``. This is the scale applied to the normalized
                inputs. This cannot be passed in if the module was constructed with
                ``create_scale=True``.
            offset: An array up to n-D. The shape of this tensor must be broadcastable
                to the shape of ``inputs``. This is the offset applied to the normalized
                inputs. This cannot be passed in if the module was constructed with
                ``create_offset=True``.

        Returns:
            The array, normalized.
        """
        if self.create_scale and scale is not None:
            raise ValueError("Cannot pass `scale` at call time if `create_scale=True`.")
        if self.create_offset and offset is not None:
            raise ValueError(
                "Cannot pass `offset` at call time if `create_offset=True`."
            )

        axis = self.axis
        if isinstance(axis, slice):
            axis = tuple(range(inputs.ndim)[axis])

        mean = jnp.mean(inputs, axis=axis, keepdims=True)
        variance = jnp.var(inputs, axis=axis, keepdims=True)

        # param_shape = inputs.shape[-1:]
        if self.create_scale:
            scale = self.scale
        elif scale is None:
            scale = np.array(1.0, dtype=inputs.dtype)

        if self.create_offset:
            offset = self.offset
        elif offset is None:
            offset = np.array(0.0, dtype=inputs.dtype)

        scale = jnp.broadcast_to(scale, inputs.shape)
        offset = jnp.broadcast_to(offset, inputs.shape)
        mean = jnp.broadcast_to(mean, inputs.shape)

        eps = jax.lax.convert_element_type(self.eps, variance.dtype)
        inv = scale * jax.lax.rsqrt(variance + eps)
        return inv * (inputs - mean) + offset

    def __repr__(self, info=None) -> str:
        info = {
            "num_channels": self.num_channels,
            "axis": self.axis,
            "create_scale": self.create_scale,
            "create_offset": self.create_offset,
        }
        return self._repr(info)



"""PAX package."""

from pax import experimental, nets, nn, utils
from pax._src.core import (
    Module,
    ParameterModule,
    PaxKind,
    StateModule,
    apply_mp_policy,
    assert_structure_equal,
    enable_eval_mode,
    enable_train_mode,
    freeze_parameters,
    module_and_value,
    pure,
    select_parameters,
    select_states,
    unfreeze_parameters,
    unwrap_mp_policy,
    update_parameters,
    update_states,
)
from pax._src.core.rng import next_rng_key, seed_rng_key
from pax._src.nn.dropout import dropout
from pax._src.utils import build_update_fn, grad, scan, value_and_grad

STATE = PaxKind.STATE
PARAMETER = PaxKind.PARAMETER
P = PaxKind.PARAMETER
S = PaxKind.STATE

__version__ = "0.4.3"

__all__ = (
    "apply_mp_policy",
    "assert_structure_equal",
    "build_update_fn",
    "dropout",
    "enable_eval_mode",
    "enable_train_mode",
    "experimental",
    "freeze_parameters",
    "grad",
    "module_and_value",
    "Module",
    "nets",
    "next_rng_key",
    "nn",
    "ParameterModule",
    "PaxKind",
    "pure",
    "scan",
    "seed_rng_key",
    "select_parameters",
    "select_states",
    "StateModule",
    "unfreeze_parameters",
    "unwrap_mp_policy",
    "update_parameters",
    "update_states",
    "utils",
    "value_and_grad",
)

try:
    del _src  # pylint: disable=undefined-variable
except NameError:
    pass



"""Transformer self-attention module."""

# Source: https://github.com/deepmind/dm-haiku/blob/main/haiku/_src/attention.py
from typing import Optional

import jax
import jax.numpy as jnp
import numpy as np

from ..core import Module
from .linear import Linear


class MultiHeadAttention(Module):
    """Multi-headed attention mechanism.
    As described in the vanilla Transformer paper:
    "Attention is all you need" https://arxiv.org/abs/1706.03762
    """

    num_heads: int
    key_size: int
    value_size: int
    model_size: int

    def __init__(
        self,
        num_heads: int,
        key_size: int,
        w_init_scale: float,
    ):
        super().__init__()
        self.num_heads = num_heads
        self.key_size = key_size
        self.value_size = key_size
        self.model_size = key_size * num_heads
        w_init = jax.nn.initializers.variance_scaling(
            w_init_scale, mode="fan_in", distribution="normal"
        )
        self.query_projection = Linear(
            self.model_size, self.model_size, w_init=w_init, name="qry_proj"
        )
        self.key_projection = Linear(
            self.model_size, self.model_size, w_init=w_init, name="key_proj"
        )
        self.value_projection = Linear(
            self.model_size, self.model_size, w_init=w_init, name="val_proj"
        )
        self.output_projection = Linear(
            self.model_size, self.model_size, w_init=w_init, name="out_proj"
        )

    def __call__(
        self,
        query: jnp.ndarray,
        key: jnp.ndarray,
        value: jnp.ndarray,
        mask: Optional[jnp.ndarray] = None,
    ) -> jnp.ndarray:
        """Compute (optionally masked) MHA with queries, keys & values."""

        query_heads = self.query_projection(query)
        key_heads = self.key_projection(key)
        value_heads = self.value_projection(value)
        (query_heads, key_heads, value_heads) = jax.tree_map(
            lambda x, y: x.reshape(*y.shape[:-1], self.num_heads, self.key_size),
            (query_heads, key_heads, value_heads),
            (query, key, value),
        )

        attn_logits = jnp.einsum("...thd,...Thd->...htT", query_heads, key_heads)
        sqrt_key_size = np.sqrt(self.key_size).astype(key.dtype)
        attn_logits = attn_logits / sqrt_key_size
        if mask is not None:
            # assert mask.shape == attn_logits.shape
            attn_logits = jnp.where(mask, attn_logits, -1e30)

        attn_weights = jax.nn.softmax(attn_logits)
        attn = jnp.einsum("...htT,...Thd->...thd", attn_weights, value_heads)
        # Concatenate attention matrix of all heads into a single vector.
        attn_vec = jnp.reshape(attn, (*query.shape[:-1], -1))
        return self.output_projection(attn_vec)

    def __repr__(self, info=None) -> str:
        info = {"num_heads": self.num_heads, "key_size": self.key_size}
        return self._repr(info)



"""Conv modules."""

# Following the jax convolution tutorial:
# https://jax.readthedocs.io/en/latest/notebooks/convolutions.html
#

from typing import Callable, Optional, Sequence, Tuple, Union

import jax
import jax.numpy as jnp
import numpy as np

from ..core import ParameterModule
from ..core.rng import KeyArray, next_rng_key


class Conv(ParameterModule):
    """Convolution Base Class."""

    weight: jnp.ndarray
    bias: Optional[jnp.ndarray]

    def __init__(
        self,
        in_features: int,
        out_features: int,
        kernel_shape: Union[int, Sequence[int]],
        stride: Union[int, Sequence[int]] = 1,
        rate: Union[int, Sequence[int]] = 1,
        padding: Union[str, Sequence[Tuple[int, int]]] = "SAME",
        with_bias: bool = True,
        w_init: Optional[Callable] = None,
        b_init: Optional[Callable] = None,
        data_format=None,
        *,
        name: Optional[str] = None,
        rng_key: Optional[KeyArray] = None,
    ):
        assert in_features > 0 and out_features > 0, "positive values"
        assert data_format in [
            "NWC",
            "NCW",
            "NHWC",
            "NCHW",
        ], f"Data format `{data_format}` is not supported."
        super().__init__(name=name)

        self.in_features = in_features
        self.out_features = out_features

        ndim = len(data_format) - 2

        if isinstance(kernel_shape, int):
            kernel_shape = (kernel_shape,) * ndim
        self.kernel_shape = kernel_shape

        if isinstance(stride, int):
            stride = (stride,) * ndim
        self.stride = stride

        if isinstance(rate, int):
            rate = (rate,) * ndim
        self.rate = rate

        if isinstance(padding, str):
            assert padding in ["SAME", "VALID"], f"Not supported padding `{padding}`"
        elif isinstance(padding, tuple):
            raise ValueError(
                "Tuple type padding is not supported. Use `[ (int, int) ]` instead."
            )
        self.padding = padding

        self.with_bias = with_bias
        self.data_format = data_format

        rng_key = next_rng_key() if rng_key is None else rng_key
        w_rng_key, b_rng_key = jax.random.split(rng_key)

        w_shape = [*kernel_shape, in_features, out_features]
        if ndim == 1:
            self.kernel_format = "WIO"
        else:
            self.kernel_format = "HWIO"
        self.kernel_dilation = (1,) * ndim

        if w_init is None:
            fan_in = np.prod(w_shape[:-1])
            w_init = jax.nn.initializers.normal(stddev=1.0 / np.sqrt(fan_in))

        self.weight = w_init(w_rng_key, w_shape)

        if with_bias:
            if b_init is None:
                b_init = jax.nn.initializers.zeros
            b_shape = [out_features]
            self.bias = b_init(b_rng_key, b_shape)
        else:
            self.bias = None

    def __call__(self, x: jnp.ndarray) -> jnp.ndarray:
        assert len(x.shape) == len(self.kernel_format)
        n_channels = x.shape[1] if self.data_format[1] == "C" else x.shape[-1]
        if n_channels != self.in_features:
            raise ValueError(
                f"Expecting {self.in_features} input channels. Get {n_channels} channels."
            )

        dimension_numbers = jax.lax.conv_dimension_numbers(
            x.shape,
            self.weight.shape,
            (self.data_format, self.kernel_format, self.data_format),
        )

        x = jax.lax.conv_general_dilated(
            x,
            self.weight,
            self.stride,
            self.padding,
            self.kernel_dilation,  # lhs/image dilation
            self.rate,  # rhs/kernel dilation
            dimension_numbers,
        )

        if self.bias is not None:
            if self.data_format == "NCHW":
                # pylint: disable=unsubscriptable-object
                x = x + self.bias[None, :, None, None]
            elif self.data_format == "NCW":
                # pylint: disable=unsubscriptable-object
                x = x + self.bias[None, :, None]
            elif self.data_format[-1] == "C":
                x = x + self.bias
            else:
                raise ValueError(
                    f"Not expecting this to happen, data_format {self.data_format}"
                )

        return x

    def __repr__(self):
        info = {
            "in_features": self.in_features,
            "out_features": self.out_features,
            "kernel_shape": self.kernel_shape,
            "padding": self.padding,
            "stride": self.stride,
            "rate": self.rate,
            "with_bias": self.with_bias,
            "data_format": self.data_format,
        }
        all_one = lambda x: all(e == 1 for e in x)
        if all_one(info["rate"]):
            del info["rate"]
        if all_one(info["stride"]):
            del info["stride"]

        return self._repr(info)


class Conv1D(Conv):
    """1D Convolution Module."""

    def __init__(
        self,
        in_features: int,
        out_features: int,
        kernel_shape: Union[int, Sequence[int]],
        stride: Union[int, Sequence[int]] = 1,
        rate: Union[int, Sequence[int]] = 1,
        padding: Union[str, Sequence[Tuple[int, int]]] = "SAME",
        with_bias: bool = True,
        w_init: Optional[Callable] = None,
        b_init: Optional[Callable] = None,
        data_format: str = "NWC",
        *,
        name: Optional[str] = None,
        rng_key: Optional[KeyArray] = None,
    ):
        """Initializes the module.

        (Haiku documentation)

        Arguments:
            in_features: Number of input channels.
            out_features: Number of output channels.
            kernel_shape: The shape of the kernel. Either an integer or a sequence of
                length 1.
            stride: Optional stride for the kernel. Either an integer or a sequence of
                length 1. Defaults to 1.
            rate: Optional kernel dilation rate. Either an integer or a sequence of
                length 1. 1 corresponds to standard ND convolution,
                ``rate > 1`` corresponds to dilated convolution. Defaults to 1.
            padding: Optional padding. Either ``VALID`` or ``SAME`` or
                sequence of `Tuple[int, int]` representing the padding before and after
                for each spatial dimension. Defaults to ``SAME``.
            with_bias: Whether to add a bias. By default, true.
            w_init: Optional weight initialization. By default, truncated normal.
            b_init: Optional bias initialization. By default, zeros.
            data_format: The data format of the input. Either ``NWC`` or ``NCW``. By
                default, ``NWC``.
            name: The name of the module.
            rng_key: The random key.
        """

        assert data_format in [
            "NWC",
            "NCW",
        ], f"Data format `{data_format}` is not supported. Use `NWC` or `NCW`."
        super().__init__(
            in_features=in_features,
            out_features=out_features,
            kernel_shape=kernel_shape,
            stride=stride,
            rate=rate,
            padding=padding,
            with_bias=with_bias,
            w_init=w_init,
            b_init=b_init,
            data_format=data_format,
            name=name,
            rng_key=rng_key,
        )


class Conv2D(Conv):
    """2D Convolution Module."""

    def __init__(
        self,
        in_features: int,
        out_features: int,
        kernel_shape: Union[int, Sequence[int]],
        stride: Union[int, Sequence[int]] = 1,
        rate: Union[int, Sequence[int]] = 1,
        padding: Union[str, Sequence[Tuple[int, int]]] = "SAME",
        with_bias: bool = True,
        w_init: Optional[Callable] = None,
        b_init: Optional[Callable] = None,
        data_format: str = "NHWC",
        *,
        name: Optional[str] = None,
        rng_key: Optional[KeyArray] = None,
    ):
        """Initializes the module.

        (Haiku documentation)

        Arguments:
            in_features: Number of output channels.
            out_features: Number of output channels.
            kernel_shape: The shape of the kernel. Either an integer or a sequence of
                length 2.
            stride: Optional stride for the kernel. Either an integer or a sequence of
                length 2. Defaults to 1.
            rate: Optional kernel dilation rate. Either an integer or a sequence of
                length 2. 1 corresponds to standard ND convolution,
                ``rate > 1`` corresponds to dilated convolution. Defaults to 1.
            padding: Optional padding. Either ``VALID`` or ``SAME`` or
                sequence of `Tuple[int, int]` representing the padding before and after
                for each spatial dimension. Defaults to ``SAME``.
            with_bias: Whether to add a bias. By default, true.
            w_init: Optional weight initialization. By default, truncated normal.
            b_init: Optional bias initialization. By default, zeros.
            data_format: The data format of the input. Either ``NHWC`` or ``NCHW``. By
                default, ``NHWC``.
            name: The name of the module.
            rng_key: The random key.
        """

        assert data_format in [
            "NHWC",
            "NCHW",
        ], f"Data format `{data_format}` is not supported. Use `NHWC` or `NCHW`."
        super().__init__(
            in_features=in_features,
            out_features=out_features,
            kernel_shape=kernel_shape,
            stride=stride,
            rate=rate,
            padding=padding,
            with_bias=with_bias,
            w_init=w_init,
            b_init=b_init,
            data_format=data_format,
            name=name,
            rng_key=rng_key,
        )


class ConvTranspose(ParameterModule):
    """Convolution Transpose Base Class."""

    weight: jnp.ndarray
    bias: jnp.ndarray

    def __init__(
        self,
        in_features: int,
        out_features: int,
        kernel_shape: Union[int, Sequence[int]],
        stride: Union[int, Sequence[int]] = 1,
        padding: Union[str, Sequence[Tuple[int, int]]] = "SAME",
        with_bias: bool = True,
        w_init: Optional[Callable] = None,
        b_init: Optional[Callable] = None,
        data_format=None,
        *,
        name: Optional[str] = None,
        rng_key: Optional[KeyArray] = None,
    ):
        assert in_features > 0 and out_features > 0, "positive values"
        assert data_format in [
            "NWC",
            "NCW",
            "NHWC",
            "NCHW",
        ], f"Data format `{data_format}` is not supported."
        super().__init__(name=name)

        self.in_features = in_features
        self.out_features = out_features

        ndim = len(data_format) - 2

        if isinstance(kernel_shape, int):
            kernel_shape = (kernel_shape,) * ndim
        self.kernel_shape = kernel_shape

        if isinstance(stride, int):
            stride = (stride,) * ndim
        self.stride = stride

        if isinstance(padding, str):
            assert padding in ["SAME", "VALID"], f"Not supported padding `{padding}`"
        self.padding = padding

        self.with_bias = with_bias
        self.data_format = data_format

        rng_key = next_rng_key() if rng_key is None else rng_key
        w_rng_key, b_rng_key = jax.random.split(rng_key)

        w_shape = [*kernel_shape, out_features, in_features]
        if ndim == 1:
            self.kernel_format = "WOI"
        else:
            self.kernel_format = "HWOI"
        self.kernel_dilation = (1,) * ndim

        if w_init is None:
            fan_in = np.prod(w_shape[:-2] + [in_features])
            w_init = jax.nn.initializers.normal(stddev=1.0 / np.sqrt(fan_in))
        if b_init is None:
            b_init = jax.nn.initializers.zeros

        self.weight = w_init(w_rng_key, w_shape)
        b_shape = [out_features]
        self.bias = b_init(b_rng_key, b_shape)

    def __call__(self, x: jnp.ndarray) -> jnp.ndarray:
        assert len(x.shape) == len(self.kernel_format)

        dimension_numbers = jax.lax.conv_dimension_numbers(
            x.shape,
            self.weight.shape,
            (self.data_format, self.kernel_format, self.data_format),
        )

        x = jax.lax.conv_transpose(
            lhs=x,
            rhs=self.weight,
            strides=self.stride,
            padding=self.padding,
            dimension_numbers=dimension_numbers,
        )

        if self.with_bias:
            if self.data_format == "NCHW":
                x = x + self.bias[None, :, None, None]
            elif self.data_format == "NCW":
                x = x + self.bias[None, :, None]
            elif self.data_format[-1] == "C":
                x = x + self.bias
            else:
                raise ValueError(
                    f"Not expecting this to happen, data_format={self.data_format}"
                )

        return x

    def __repr__(self):
        info = {
            "in_features": self.in_features,
            "out_features": self.out_features,
            "data_format": self.data_format,
            "padding": self.padding,
            "stride": self.stride,
            "with_bias": self.with_bias,
        }
        return self._repr(info)


class Conv1DTranspose(ConvTranspose):
    """1D Convolution Transpose Module."""

    def __init__(
        self,
        in_features: int,
        out_features: int,
        kernel_shape: Union[int, Sequence[int]],
        stride: Union[int, Sequence[int]] = 1,
        padding: Union[str, Sequence[Tuple[int, int]]] = "SAME",
        with_bias: bool = True,
        w_init: Optional[Callable] = None,
        b_init: Optional[Callable] = None,
        data_format: str = "NWC",
        *,
        name: Optional[str] = None,
        rng_key: Optional[KeyArray] = None,
    ):
        """Initializes the module.

        (Haiku documentation)


        Arguments:
            in_features: Number of input channels.
            out_features: Number of output channels.
            kernel_shape: The shape of the kernel. Either an integer or a sequence of
                length 1.
            stride: Optional stride for the kernel. Either an integer or a sequence of
                length 1. Defaults to 1.
            padding: Optional padding algorithm. Either ``VALID`` or ``SAME``.
                Defaults to ``SAME``.
            with_bias: Whether to add a bias. By default, true.
            w_init: Optional weight initialization. By default, truncated normal.
            b_init: Optional bias initialization. By default, zeros.
            data_format: The data format of the input. Either ``NWC`` or ``NCW``. By
                default, ``NWC``.
            name: The name of the module.
            rng_key: The random key.
        """

        assert data_format in [
            "NWC",
            "NCW",
        ], f"Data format `{data_format}` is not supported. Use `NWC` or `NCW`."

        super().__init__(
            in_features=in_features,
            out_features=out_features,
            kernel_shape=kernel_shape,
            stride=stride,
            padding=padding,
            with_bias=with_bias,
            w_init=w_init,
            b_init=b_init,
            data_format=data_format,
            name=name,
            rng_key=rng_key,
        )


class Conv2DTranspose(ConvTranspose):
    """2D Convolution Transpose Module."""

    def __init__(
        self,
        in_features: int,
        out_features: int,
        kernel_shape: Union[int, Sequence[int]],
        stride: Union[int, Sequence[int]] = 1,
        padding: Union[str, Sequence[Tuple[int, int]]] = "SAME",
        with_bias: bool = True,
        w_init: Optional[Callable] = None,
        b_init: Optional[Callable] = None,
        data_format: str = "NHWC",
        *,
        name: Optional[str] = None,
        rng_key: Optional[KeyArray] = None,
    ):
        """Initializes the module.

        (Haiku documentation)


        Arguments:
            in_features: Number of input channels.
            out_features: Number of output channels.
            kernel_shape: The shape of the kernel. Either an integer or a sequence of
                length 1.
            stride: Optional stride for the kernel. Either an integer or a sequence of
                length 1. Defaults to 1.
            padding: Optional padding algorithm. Either ``VALID`` or ``SAME``.
                Defaults to ``SAME``.
            with_bias: Whether to add a bias. By default, true.
            w_init: Optional weight initialization. By default, truncated normal.
            b_init: Optional bias initialization. By default, zeros.
            data_format: The data format of the input. Either ``NHWC`` or ``NHCW``. By
                default, ``NHWC``.
            name: The name of the module.
            rng_key: The random key.
        """

        assert data_format in [
            "NHWC",
            "NCHW",
        ], f"Data format `{data_format}` is not supported. Use `NHWC` or `NCHW`."

        super().__init__(
            in_features=in_features,
            out_features=out_features,
            kernel_shape=kernel_shape,
            stride=stride,
            padding=padding,
            with_bias=with_bias,
            w_init=w_init,
            b_init=b_init,
            data_format=data_format,
            name=name,
            rng_key=rng_key,
        )
